import re

import xarray as xr
import cf_units


@xr.register_dataarray_accessor('cf')
class CFDataArray:
    def __init__(self, da: xr.Dataset):
        self.da = da

    @property
    def attrs(self) -> dict:
        return self.da.attrs

    @property
    def units(self)-> cf_units.Unit:
        if 'units' in self.attrs:
            return cf_units.Unit(self.attrs['units'])
        else:
            return None

    @units.setter
    def units(self, units):
        if units is None and 'units' in self.attrs:
            del self.attrs['units']
        if type(units) == cf_units.Unit:
            self.calendar = units.calendar
        self.attrs['units'] = str(units)

    @property
    def calendar(self):
        return self.attrs.get('calendar', None)

    @calendar.setter
    def calendar(self, calendar):
        if calendar is None and 'calendar' in self.attrs:
            del self.attrs['calendar']
        elif calendar is None or calendar == "none":
            if self.calendar is not None:
                self.attrs['calendar'] = "none"
        else:
            self.attrs['calendar'] = calendar

    @property
    def long_name(self) -> str:
        return self.attrs.get('long_name', None)

    @long_name.setter
    def long_name(self, long_name):
        if long_name is None and 'long_name' in self.attrs:
            del self.attrs['long_name']
        self.attrs['long_name'] = long_name

    @property
    def standard_name(self) -> str:
        return self.attrs.get('standard_name', None)

    @standard_name.setter
    def standard_name(self, standard_name):
        if standard_name is None and 'standard_name' in self.attrs:
            del self.attrs['standard_name']
        self.attrs['standard_name'] = standard_name

    @property
    def positive(self) -> str:
        return self.attrs.get('positive', None)

    @positive.setter
    def positive(self, positive: str):
        if positive is None and 'positive' in self.attrs:
            del self.attrs['positive']
        self.attrs['positive'] = positive

    @property
    def formula_terms(self) -> dict:
        if 'formula_terms' in self.attrs:
            regex = re.compile(r'(\w+):\s?(\w+)')
            groups = regex.findall(self.attrs['formula_terms'])
            terms = {g[0]: g[1] for g in groups}
            return terms
        else:
            return None

    @formula_terms.setter
    def formula_terms(self, terms: dict):
        if terms is None and 'formula_terms' in self.attrs:
            del self.attrs['formula_terms']
        formula_terms = ''
        for term_name, term_var in terms.items():
            formula_terms += f'{term_name}: {term_var} '
        self.attrs['formula_terms'] = formula_terms.strip()

    def convert_units_to(self, target_units: str, target_calendar=None):
        target = cf_units.Unit(target_units, target_calendar)
        new = self.da.copy().cf
        if target == self.units:
            return new
        new.da.values = self.units.convert(self.da.values, target)
        new.units = target
        if new.da.name in new.da.coords:
            new.da.coords[new.da.name] = new.da
        new.standard_name = None
        return new.da


@xr.register_dataset_accessor('cf')
class CFDataset:
    def __init__(self, ds: xr.Dataset):
        self.ds = ds

    def find_long_name(self, long_name: str) -> xr.DataArray:
        for var in self.ds.variables:
            da = self.ds[var]
            if da.attrs.get('long_name', None) == long_name:
                return da
        raise KeyError(f'long_name {long_name} is not defined')

    def find_standard_name(self, standard_name: str) -> xr.DataArray:
        for var in self.ds.variables:
            da = self.ds[var]
            if da.attrs.get('standard_name', None) == standard_name:
                return da
        raise KeyError(f'standard_name {standard_name} is not defined')

    def __getitem__(self, name):
        try:
            return self.find_standard_name(name)
        except KeyError:
            pass
        try:
            return self.find_long_name(name)
        except KeyError:
            pass
        try:
            return self.ds[name]
        except KeyError:
            pass
        raise KeyError(f"Variable '{name}' not found.")

    def __setitem__(self, key, value):
        try:
            self.ds[self[key].name] = value
        except KeyError:
            self.ds[key] = value
            self.ds[key].cf.long_name = key


if __name__ == '__main__':
    ds = xr.open_dataset('/home/liam/nested1/nested2/cache/aligned-surface_parameters-tempo_sampled-2013-12-21-18.nc')
    lat = ds.latitude.cf
    assert lat.units is None
    assert lat.calendar is None
    assert lat.long_name is None
    assert lat.standard_name is None
    assert lat.positive is None
    assert lat.formula_terms is None
    lat.units = 'm'
    assert ds.latitude.attrs['units'] == 'm'
    assert type(lat.units) == cf_units.Unit and str(lat.units) == 'm'
    ds['latitude'] = lat.da
    assert 'units' in ds.latitude.attrs
    inches = lat.convert_units_to('in').cf
    assert type(lat.units) == cf_units.Unit and str(lat.units) == 'm'
    assert type(inches.units) == cf_units.Unit and str(inches.units) == 'in'
