import argparse
import numpy as np
import xarray as xr


def derivative_parser():
    parser = argparse.ArgumentParser(
        description='Estimates derivatives of an AMF table. Derivatives are taken with respect to albedo for'
                    'clear AMFs and cloud pressure for cloudy AMFs')
    parser.add_argument('filename', type=str, help='path to file AMF table')
    parser.add_argument('dataarray', type=str, help='name of the DataArray containing AMFs')
    parser.add_argument('coord', type=str, help='name of the coordinate to take derivatives with respect to')
    return parser


def derivative_operator(x: np.ndarray):
    n = len(x)
    if n < 3:
        raise ValueError('x must have a length of at least 3')
    d = x[1:] - x[:-1]
    if np.min(d) <= 0:
        raise ValueError('x must be increasing')

    M = np.zeros((n, n))
    M[0, 0] = -(2 * d[0] + d[1]) / (d[0] * (d[0] + d[1]))
    M[0, 1] = (d[0] + d[1]) / (d[0] * d[1])
    M[0, 2] = -d[0] / (d[1] * (d[0] + d[1]))

    for i in range(1, n - 1):
        M[i, i - 1] = -d[i] / (d[i - 1] * (d[i - 1] + d[i]))
        M[i, i] = -(d[i - 1] - d[i]) / (d[i - 1] * d[i])
        M[i, i + 1] = d[i - 1] / (d[i] * (d[i - 1] + d[i]))

    M[-1, -3] = d[-1] / (d[-2] * (d[-2] + d[-1]))
    M[-1, -2] = -(d[-2] + d[-1]) / (d[-2] * d[-1])
    M[-1, -1] = (d[-2] + 2 * d[-1]) / (d[-1] * (d[-2] + d[-1]))

    return M


def take_derivative(filename: str, dataarray: str, coord: str):
    with xr.open_dataset(filename).load() as ds:
        dv_op = derivative_operator(ds[dataarray][coord].values)
        i = ds[dataarray].dims.index(coord)
        dv = np.swapaxes(dv_op @ np.swapaxes(ds[dataarray].values, i, -2), i, -2)
        longname = ds[dataarray].attrs['long_name'] if 'long_name' in ds[dataarray].attrs else dataarray
        longnamecoord = ds[coord].attrs['long_name'] if 'long_name' in ds[coord].attrs else coord
        ds[f'{dataarray}_{coord}'] = (ds[dataarray].dims, dv, {'long_name': f'derivative of {longname} with respect to {longnamecoord}'})
    ds.to_netcdf(filename)


if __name__ == '__main__':

    parser = derivative_parser()
    args = vars(parser.parse_args())

    # try opening the table, check dataarray and coord for consistency
    with xr.open_dataset(args['filename']) as ds:
        if args['dataarray'] not in ds:
            raise ValueError(f'DataArray {args["dataarray"]} was not found in the AMF table')
        if args['coord'] not in ds[args['dataarray']].coords:
            raise ValueError(f'Coordinate {args["coord"]} not found in DataArray {args["dataarray"]}')

    take_derivative(**args)
