"""
This program calculates AMF entries to fill an AMF table. It operates by looping through the coordinates in the AMF
table, interpreting the coordinates, and then computing the AMFs. The table's coordinates must be compliant with CF
Conventions (i.e. have a "standard_name" attribute with the corresponding CF standard name).
"""

import argparse
import os
import sys
import logging
import pickle as pkl
import signal
import time
import datetime
from uuid import uuid4
import warnings

import xarray as xr

import skamf
from skamf.iterators import ClearAtmosphereVariables, CloudyAtmosphereVariables, GeometricVariables
from skamf.amf import ClearSkyAMF, CloudySkyAMF
from skamf.skamflogging import TqdmHandler
from skamf.new import create_table

import skamf.cf


warnings.simplefilter(action='ignore', category=FutureWarning)
last_save = ''


def process_standard(amf_table, progress_file=None, log_level=None, save_period=None, engine=None, reprocess=None,
                     R=None, ignore_version=None, **kwargs):

    # setup logging
    log = logging.getLogger()
    handler = TqdmHandler()
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)-8s %(message)s')
    handler.setFormatter(formatter)
    log.addHandler(handler)
    log.setLevel(logging.getLevelName(log_level))

    # make amf_table ends in ".nc"
    split_amf_table_filename = amf_table.rsplit('.', 1)
    if len(split_amf_table_filename) < 2 or (len(split_amf_table_filename) > 1 and split_amf_table_filename[1] != 'nc'):
        print("argument amf_table must finish with '.nc'")
        exit(1)

    # print AMF version
    log.debug(f'Version: {skamf.__version__}')

    # load amf_table table
    log.info(f'Loading AMF table: {amf_table}')
    table = xr.open_dataset(amf_table).load()
    table.close()

    if engine != 'choose':
        if 'permitted_engine' in table.attrs:
            if engine != table.attrs['permitted_engine']:
                log.error(f'This table must be processed with the {table.attrs["permitted_engine"]} engine, not the '
                          f'{engine} engine')
                exit(1)
    else:
        if 'permitted_engine' not in table.attrs:
            engine = 'do'
        else:
            engine = table.attrs['permitted_engine']

    # the "date_created" attribute is present iff the AMF has been fully processed
    if 'date_created' in table.attrs or 'date_created' in table.variables:
        if reprocess or R:
            inp = ''
            if not R:
                while inp.lower() not in ['y', 'n']:
                    inp = input(f'{amf_table} will be cleared and reprocessed. Continue? [y/n] ')
            if inp.lower() == 'y' or R:
                create_table(filename=amf_table, copy=amf_table, copy_options=amf_table)
                table = xr.open_dataset(amf_table).load()
                table.close()
            else:
                exit(0)
        else:
            log.error(f'{args["amf_table"]} is fully processed')
            exit(1)

    # check that the version is not wrong
    if 'version' in table.attrs and not reprocess and not ignore_version:
        differences = _compare_dependency_versions(table.attrs['version'])
        if differences:
            errmsg = f'Inconsistent versions'
            for package, (vtable, vcurrent) in differences.items():
                errmsg += f' | {package}: {vtable} (table), {vcurrent} (current)'
            log.error(errmsg)
            exit(1)

    # get the path to the progress file
    if progress_file is not None:
        # only use p if table.attrs['progress_file'] is missing
        if 'progress_file' not in table.attrs:
            pfile = os.path.abspath(progress_file)
            table.attrs['progress_file'] = pfile
        elif table.attrs['progress_file'] == os.path.abspath(progress_file):
            pfile = table.attrs['progress_file']
        else:
            log.warning(f"Ignoring argument '-p {progress_file}'. "
                        f"The progress file for '{amf_table}' is '{table.attrs['progress_file']}'")
            pfile = table.attrs['progress_file']
    else:
        # generate filename if not present in the table
        if 'progress_file' in table.attrs:
            pfile = table.attrs['progress_file']
            if reprocess or R:
                inp = ''
                if not R:
                    while inp.lower() not in ['y', 'n']:
                        inp = input(
                            f'{amf_table} and progress file {pfile} will be cleared and reprocessed. Continue? [y/n] '
                        )
                if inp.lower() == 'y' or R:
                    create_table(filename=amf_table, copy=amf_table, copy_options=amf_table)
                    table = xr.open_dataset(amf_table).load()
                    table.close()
                    os.remove(pfile)
                else:
                    exit(0)
        if 'progress_file' not in table.attrs:
            pfile = os.path.abspath(os.path.join(split_amf_table_filename[0] + f'-{str(uuid4())[:8]}.pkl'))

        table.attrs['version'] = _get_dependency_versions()

    # define two save functions: one saves the progress file and amf_table, the other calls save() periodically
    def save():
        log.debug('Saving AMF table')
        # save AMF table
        table.to_netcdf(amf_table)
        # save progress file
        log.debug('Saving progress file')
        with open(pfile, 'wb') as file:
            pkl.dump(progress, file)
    global last_save
    last_save = time.time()

    def save_periodically(force=False, **next_indexes):
        global last_save
        index_keys = set(next_indexes.keys())
        if force or time.time() - last_save > save_period:
            for k, s in next_indexes.items():
                progress[k] = s.start
            log.debug(f"Saving checkpoint: { {k: progress[k] for k in index_keys} }")
            save()
            last_save = time.time()
        # we don't want to load the current progress so clear them
        for k in index_keys:
            progress.pop(k, None)

    # load/make the progress file
    if os.path.exists(pfile):
        log.info(f'Loading progress file: {pfile}')
        with open(pfile, 'rb') as file:
            progress = pkl.load(file)
    else:
        progress = {
            'amf_table': os.path.abspath(amf_table),
            'cloudy_sky_complete': False,
            'clear_sky_complete': False
        }
        table.attrs['progress_file'] = pfile
        log.info(f'Creating progress file: {pfile}')
        save()

    log.info(f'Autosave period: {args["save_period"]} [seconds]')

    # setup SIGINT interrupt
    def graceful_interrupt(signal, frame):
        log.info(f"User interrupted computation.")
        exit(0)
    signal.signal(signal.SIGINT, graceful_interrupt)

    try:
        tmp = table.cf['clear_box_amf']
        process_clear = True
    except KeyError:
        process_clear = False
    if process_clear and progress['clear_sky_complete'] is False:
        log.info('Processing clear-sky box AMFs.')

        # iterate through the table coordinates and calculate AMFs
        for atmos_isel in ClearAtmosphereVariables(table, **progress):
            for geometry_isel in GeometricVariables(table, **progress):
                # save periodically (this should be called BEFORE this items processing)
                save_periodically(**atmos_isel, **geometry_isel)
                subset = table.isel(**atmos_isel, **geometry_isel).squeeze()
                ctor = ClearSkyAMF(subset, engine)
                ctor.calculate()

                dims = subset.cf['clear_box_amf'].dims
                if len(dims) > 0:
                    subset.cf['clear_box_amf'][::] = ctor.box_amf.transpose(*dims).values
                else:
                    table.cf['clear_box_amf'] = ctor.box_amf.values

                if 'clear_box_amf_std' in subset:
                    dims = subset.cf['clear_box_amf_std'].dims
                    if len(dims) > 0:
                        subset.cf['clear_box_amf_std'][::] = ctor.box_amf_std.transpose(*dims).values
                    else:
                        table.cf['clear_box_amf_std'] = ctor.box_amf_std.values

                if 'clear_amf_box_sum' in subset:
                    dims = subset.cf['clear_amf_box_sum'].dims
                    if len(dims) > 0:
                        subset.cf['clear_amf_box_sum'][::] = ctor.amf_box_sum.transpose(*dims).values
                    else:
                        table.cf['clear_amf_box_sum'] = ctor.amf_box_sum.values

                if 'clear_amf_no_NO2' in subset:
                    dims = subset.cf['clear_amf_no_NO2'].dims
                    if len(dims) > 0:
                        subset.cf['clear_amf_no_NO2'][::] = ctor.amf_no_absorber.transpose(*dims).values
                    else:
                        table.cf['clear_amf_no_NO2'] = ctor.amf_no_absorber.values

    # We're done the clear sky
    progress['clear_sky_complete'] = True
    save()

    try:
        tmp = table.cf['cloudy_box_amf']
        process_cloudy = True
    except KeyError:
        process_cloudy = False
    if process_cloudy and progress['cloudy_sky_complete'] is False:
        log.info('Processing cloudy-sky box AMFs.')

        # iterate through the table coordinates and calculate AMFs
        for atmos_isel in CloudyAtmosphereVariables(table, **progress):
            for geometry_isel in GeometricVariables(table, **progress):
                # save periodically (this should be called BEFORE this items processing)
                save_periodically(**atmos_isel, **geometry_isel)
                subset = table.isel(**atmos_isel, **geometry_isel).squeeze()
                ctor = CloudySkyAMF(subset, engine)
                if ctor.calculate():
                    dims = subset.cf['cloudy_box_amf'].dims
                    if len(dims) > 0:
                        subset.cf['cloudy_box_amf'][::] = ctor.box_amf.transpose(*dims).values
                    else:
                        table.cf['cloudy_box_amf'] = ctor.box_amf.values

                    if 'cloudy_box_amf_std' in subset:
                        dims = subset.cf['cloudy_box_amf_std'].dims
                        if len(dims) > 0:
                            subset.cf['cloudy_box_amf_std'][::] = ctor.box_amf_std.transpose(*dims).values
                        else:
                            table.cf['cloudy_box_amf_std'] = ctor.box_amf_std.values

                    if 'cloudy_amf_box_sum' in subset:
                        dims = subset.cf['cloudy_amf_box_sum'].dims
                        if len(dims) > 0:
                            subset.cf['cloudy_amf_box_sum'][::] = ctor.amf_box_sum.transpose(*dims).values
                        else:
                            table.cf['cloudy_amf_box_sum'] = ctor.amf_box_sum.values

                    if 'cloudy_amf_no_NO2' in subset:
                        dims = subset.cf['cloudy_amf_no_NO2'].dims
                        if len(dims) > 0:
                            subset.cf['cloudy_amf_no_NO2'][::] = ctor.amf_no_absorber.transpose(*dims).values
                        else:
                            table.cf['cloudy_amf_no_NO2'] = ctor.amf_no_absorber.values

                    if 'pressure' in subset:
                        subset.cf['air_pressure'][::] = ctor.altitude_to_pressure(subset.cf['altitude'])

    # We're done the cloudy sky
    progress['cloudy_sky_complete'] = True

    # delete the "progress_file" attribute and add the "date_created" attribute
    del table.attrs['progress_file']
    table.attrs['date_created'] = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M')
    if engine == 'hr':
        table.attrs['radiative_transfer_engine'] = 'SASKTRAN-HR'
    if engine == 'disco':
        table.attrs['radiative_transfer_engine'] = 'SASKTRAN-DISCO'
    if engine == 'mc':
        table.attrs['radiative_transfer_engine'] = 'SASKTRAN-MC'
    if engine == 'choose':
        table.attrs['radiative_transfer_engine'] = 'SASKTRAN-DISCO'

    # always save and delete the resulting progress file
    save()
    os.remove(pfile)
    log.info('Finished generating AMF table.')


def process_mc_cumulative(amf_table, progress_file=None, log_level=None, save_period=None, engine=None, reprocess=None,
                          R=None, ignore_version=None, number=None):

    """
    Processing designed to allow for many trials to be run, so that a large number of MC photons can be simulated
    while saving intermittently. Creates temporary tables (netcdf files) and processes them using the standard process,
    then combines them with the base table.
    """
    if number is None or number < 1:
        print('--number must be at least 1 for cumulative processing')
        return

    if reprocess:
        inp = ''
        while inp.lower() not in ['y', 'n']:
            inp = ''
            while inp.lower() not in ['y', 'n']:
                inp = input(f'{amf_table} will be cleared and reprocessed. Continue? [y/n] ')
            if inp.lower() == 'y':
                create_table(filename=amf_table, copy=amf_table, copy_options=amf_table)
            else:
                exit(0)

    # open the base table
    base_table = xr.open_dataset(amf_table).load()
    base_table.close()

    if 'trials' in base_table.dims and base_table.dims['trials'] >= number:
        print(f'requested number of trials {number} must be larger than existing number {base_table.dims["trials"]}')
        exit(1)

    if 'date_created' not in base_table.attrs:  # if the file isn't processed, finish processing it the old way
        print(f'processing trial 1/{number}')
        process_standard(amf_table=amf_table, progress_file=progress_file, log_level=log_level, save_period=save_period,
                         engine=engine, reprocess=False)

    # reopen the base table
    base_table = xr.open_dataset(amf_table).load()
    base_table.close()

    non_data_attrs = ['HSP_atmosphere', 'static_atmosphere', 'sasktran_option']
    trial_arrays = [v for v in base_table.data_vars if all([a not in non_data_attrs for a in base_table[v].attrs])]
    trial_arrays = list(set(trial_arrays) - {'date_created', 'version'})

    if 'trials' not in base_table.dims:
        base_table = _add_trials_dimension(base_table, trial_arrays)

    while base_table.dims['trials'] < number:
        print(f'processing trial {base_table.dims["trials"] + 1}/{number}')

        dir = os.path.splitext(amf_table)[0]
        if not os.path.isdir(dir):
            os.mkdir(dir)
        tempfile = os.path.abspath(
            os.path.join(dir, f'{os.path.splitext(os.path.basename(amf_table))[0]}-t{base_table.dims["trials"] + 1}.nc')
        )
        create_table(filename=tempfile, copy=amf_table, copy_options=amf_table)
        process_standard(tempfile, log_level=log_level, save_period=save_period, engine=engine)

        temp_table = xr.open_dataset(tempfile).load()
        temp_table.close()

        temp_table = _add_trials_dimension(temp_table, trial_arrays)
        print('base', base_table)
        print('temp', temp_table)

        base_table = xr.concat((base_table, temp_table), dim='trials', data_vars='minimal')
        base_table.to_netcdf(amf_table)

    print(base_table)


def _add_trials_dimension(ds: xr.Dataset, dataarrays: list):
    if 'trials' not in ds.dims:
        for var in dataarrays:
            dims = ['trials'] + list(ds[var].dims)
            shape = [1] + list(ds[var].values.shape)
            ds[var] = (dims, ds[var].values.reshape(*shape))
        if 'date_created' in ds.attrs:
            ds['date_created'] = ('trials', [ds.attrs['date_created']])
        else:
            print('base_table should have a date_created attribute')
            exit(1)
        if 'version' in ds.attrs:
            version = (ds.attrs['version'] + ' ' * 32)[:32]
            ds['version'] = ('trials', [version])
        else:
            print('the amf table should have a version attribute')
            exit(1)
    return ds


def _get_dependency_versions():
    required_packages = ['skamf', 'cf-units', 'cftime', 'numpy', 'Cython', 'six', 'sasktran', 'appdirs', 'PyYAML',
                         'sasktranif', 'sktran-disco', 'xarray', 'pandas', 'python-dateutil', 'pytz', 'doas']
    package_versions = []
    from pip._internal.operations.freeze import freeze
    for x in freeze():
        if x.split('==')[0] in required_packages:
            package_versions.append(x)
    return '\n'.join(package_versions)


def _compare_dependency_versions(old_dependency_versions):
    new_dependency_versions = _get_dependency_versions()

    old = dict([tuple(line.split('==')) for line in old_dependency_versions.split('\n')])
    new = dict([tuple(line.split('==')) for line in new_dependency_versions.split('\n')])

    differences = {}
    required_packages = ['skamf', 'cf-units', 'cftime', 'numpy', 'Cython', 'six', 'sasktran', 'appdirs', 'PyYAML',
                         'sasktranif', 'sktran-disco', 'xarray', 'pandas', 'python-dateutil', 'pytz', 'doas']
    for package in required_packages:
        if old[package] != new[package]:
            differences[package] = (old[package], new[package])

    return differences


def process_parser():
    parser = argparse.ArgumentParser(description=f"Processes an AMF table. This program generates AMF's using the "
                                     f"coordinates present in the table. Version: {skamf.__version__}")
    parser.add_argument('amf_table', type=str, help="path to the AMF table")
    parser.add_argument('-p', '--progress-file', type=str, help='name of the progress file')
    parser.add_argument('-l', '--log-level', type=str, help='logging level (default: INFO)', default='INFO')
    parser.add_argument('-s', '--save-period', type=int, default=300,
                        help='number of seconds between saves (default: 300)')
    parser.add_argument('-e', '--engine', type=str, choices=['hr', 'do', 'mc', 'choose'], default='choose',
                        help='specify which engine to use: hr, disco or choose (default: choose)')
    parser.add_argument('-r', '--reprocess', action='store_true', help='reprocess a completed table')
    parser.add_argument('-R', action='store_true', help='reprocess a completed table without confirmation')
    parser.add_argument('-n', '--number', type=int, help='number of times to process table in mc cumulative mode')
    parser.add_argument('-v', '--ignore-version', action='store_true', help='proceed even with different versions')
    return parser


if __name__ == '__main__':
    parser = process_parser()

    if len(sys.argv) > 1:
        args = vars(parser.parse_args())  # command line access
    else:
        args = vars(parser.parse_args(['path-to-table', 'args']))  # IDE access

    if args['number'] is None:
        process_standard(**args)
    else:
        process_mc_cumulative(**args)
